from dataclasses import dataclass
from typing import Optional
from unittest.mock import ANY

from serpyco_rs import Serializer
from serpyco_rs._describe import (
    EntityField,
    EntityType,
    OptionalType,
    RecursionHolder,
    StringType,
    describe_type,
)
from serpyco_rs._impl import DefaultValue


@dataclass
class Node:
    value: str
    next: Optional['Node'] = None


@dataclass
class Root:
    head: Node


def test_describe__recursive_type__parsed():
    assert describe_type(Root) == EntityType(
        cls=Root,
        name=ANY,
        fields=[
            EntityField(
                name='head',
                dict_key='head',
                field_type=EntityType(
                    cls=Node,
                    name=ANY,
                    fields=[
                        EntityField(
                            name='value',
                            dict_key='value',
                            field_type=StringType(custom_encoder=None),
                        ),
                        EntityField(
                            name='next',
                            dict_key='next',
                            default=DefaultValue.some(None),
                            required=False,
                            field_type=OptionalType(
                                inner=RecursionHolder(
                                    name=ANY,
                                    state_key=ANY,
                                    meta=ANY,
                                    custom_encoder=None,
                                ),
                                custom_encoder=None,
                            ),
                        ),
                    ],
                    doc=ANY,
                    custom_encoder=None,
                ),
            )
        ],
        doc=ANY,
        custom_encoder=None,
    )


def test_serializer():
    serializer = Serializer(Root)

    linked_list = Root(
        head=Node(
            value='1',
            next=Node(value='2'),
        ),
    )

    assert serializer.dump(linked_list) == {'head': {'next': {'next': None, 'value': '2'}, 'value': '1'}}
    assert serializer.load({'head': {'next': {'next': None, 'value': '2'}, 'value': '1'}}) == linked_list


@dataclass
class Foo:
    value: str
    next: Optional[list['Foo']] = None


def test_self_recursive_objects_forward_ref():
    serializer = Serializer(Foo)
    val = Foo(value='a', next=[Foo(value='b')])
    raw = {'value': 'a', 'next': [{'next': None, 'value': 'b'}]}
    assert serializer.dump(val) == raw
    assert serializer.load(raw) == val



import sys
import uuid
from dataclasses import dataclass
from datetime import date, datetime, time, timedelta, timezone
from decimal import Decimal
from enum import Enum, IntEnum
from typing import Generic, Literal, TypeVar
from zoneinfo import ZoneInfo

import pytest
from dateutil.tz import tzoffset
from serpyco_rs import Serializer, ValidationError
from typing_extensions import TypedDict


@pytest.mark.parametrize(
    ['type', 'value'],
    (
        (int, 44),
        (str, '123'),
        (float, 4.3),
        (bool, True),
    ),
)
def test_simple_types(type, value):
    serializer = Serializer(type)
    dump_result = serializer.dump(value)
    assert serializer.load(dump_result) == value


def test_decimal():
    serializer = Serializer(Decimal)
    assert serializer.dump(Decimal('123.1')) == '123.1'
    assert serializer.load(123.1) == Decimal('123.1')
    assert serializer.load('123.1') == Decimal('123.1')


def test_decimal_invalid_value__raise_validation_error():
    serializer = Serializer(Decimal)

    with pytest.raises(ValidationError):
        serializer.load('asd')


def test_dict_encoder():
    serializer = Serializer(dict[str, Decimal])
    val = {'a': Decimal('123.3')}
    assert serializer.dump(val) == {'a': '123.3'}
    assert serializer.load({'a': '123.3'}) == val


def test_array_encoder():
    serializer = Serializer(list[int])
    val = [1, 2, 3]
    assert serializer.dump(val) == serializer.load(val) == val


def test_entity_encoder():
    @dataclass
    class A:
        int_f: int
        float_f: float
        bool_f: bool
        str_f: str

    serializer = Serializer(A)

    obj = A(
        int_f=123,
        float_f=3.14,
        bool_f=True,
        str_f='Test',
    )
    expected = {'bool_f': True, 'float_f': 3.14, 'int_f': 123, 'str_f': 'Test'}
    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_uuid():
    serializer = Serializer(uuid.UUID)
    val = uuid.uuid4()
    assert serializer.dump(val) == str(val)
    assert serializer.load(str(val)) == val


def test_enum():
    class Foo(Enum):
        foo = 'foo'

    serializer = Serializer(Foo)
    assert serializer.dump(Foo.foo) == 'foo'
    assert serializer.load('foo') == Foo.foo


def test_tuple():
    serializer = Serializer(tuple[int, bool, str])
    assert serializer.dump((1, True, 's')) == [1, True, 's']
    assert serializer.load([1, True, 's']) == (1, True, 's')


@pytest.mark.parametrize(
    ['value', 'expected'],
    (
        (datetime(2022, 10, 10, 14, 23, 43), '2022-10-10T14:23:43'),
        (datetime(2022, 10, 10, 14, 23, 43, 123456), '2022-10-10T14:23:43.123456'),
        (
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone.utc),
            '2022-10-10T14:23:43Z',
        ),
        (
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone(timedelta(hours=1))),
            '2022-10-10T14:23:43+01:00',
        ),
        (
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=ZoneInfo('Europe/Berlin')),
            '2022-10-10T14:23:43+02:00',
        ),
    ),
)
def test_datetime_dump(value, expected):
    serializer = Serializer(datetime)
    assert serializer.dump(value) == expected


def test_datetime_dump__naive_datetime_to_utc__expect_utc():
    serializer = Serializer(datetime, naive_datetime_to_utc=True)
    assert serializer.dump(datetime(2022, 10, 10, 14, 23, 43)) == '2022-10-10T14:23:43Z'


@pytest.mark.parametrize(
    ['value', 'expected'],
    (
        ('2022-10-10T14:23:43', datetime(2022, 10, 10, 14, 23, 43)),
        ('2022-10-10T14:23:43.123456', datetime(2022, 10, 10, 14, 23, 43, 123456)),
        (
            '2022-10-10T14:23:43+00:00',
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone.utc),
        ),
        (
            '2022-10-10T14:23:43Z',
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone.utc),
        ),
        (
            '2022-10-10T14:23:43+01:00',
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=timezone(timedelta(hours=1))),
        ),
        (
            '2022-10-10T14:23:43+02:00',
            datetime(2022, 10, 10, 14, 23, 43, tzinfo=ZoneInfo('Europe/Berlin')),
        ),
        (
            '2024-04-02T12:21:53.725421224',
            datetime(2024, 4, 2, 12, 21, 53, 725421),
        ),
    ),
)
def test_datetime_load(value, expected):
    serializer = Serializer(datetime)
    assert serializer.load(value) == expected


@pytest.mark.parametrize(
    ['value', 'expected'],
    [
        ('12:34', time(12, 34)),
        ('12:34:56', time(12, 34, 56)),
        ('12:34:56.000078', time(12, 34, 56, 78)),
        ('12:34:57.000095987', time(12, 34, 57, 95)),
        (
            '12:34:56.000078+03:00',
            time(12, 34, 56, 78, tzinfo=tzoffset(None, timedelta(hours=3))),
        ),
    ],
)
def test_time_load(value, expected):
    serializer = Serializer(time)
    assert serializer.load(value) == expected


@pytest.mark.parametrize(
    ['value', 'expected'],
    [
        (time(12, 34), '12:34:00'),
        (time(12, 34, tzinfo=tzoffset(None, 10800)), '12:34:00+03:00'),
        (time(12, 34, 56), '12:34:56'),
        (time(12, 34, 56, 78), '12:34:56.000078'),
        (time(12, 34, tzinfo=timezone.utc), '12:34:00Z'),
        (time(12, 34, tzinfo=timezone(timedelta(hours=1))), '12:34:00+01:00'),
    ],
)
def test_time_dump(value, expected):
    serializer = Serializer(time)
    assert serializer.dump(value) == expected


def test_date():
    serializer = Serializer(date)
    assert serializer.load('2022-10-14') == date(2022, 10, 14)
    assert serializer.dump(date(2022, 10, 13)) == '2022-10-13'


def test_date__dump_datatime__expect_date():
    serializer = Serializer(date)
    assert serializer.dump(datetime(2022, 10, 13, 12, 34, 56)) == '2022-10-13'


def test_literal():
    serializer = Serializer(Literal['foo', 'bar'])
    assert serializer.load('bar') == 'bar'
    assert serializer.dump('foo') == 'foo'


@pytest.mark.skipif(sys.version_info < (3, 10), reason='New style unions available after 3.10')
def test_optional():
    @dataclass
    class T:
        foo: int | None = None

    serializer = Serializer(T)
    assert serializer.dump(T()) == {'foo': None}
    assert serializer.dump(T(foo=1)) == {'foo': 1}
    assert serializer.load({}) == T()
    assert serializer.load({'foo': 12}) == T(foo=12)


def test_int_enum():
    class Foo(IntEnum):
        foo = 1
        bar = 2

    serializer = Serializer(Foo)
    assert serializer.dump(Foo.foo) == 1
    assert serializer.load(1) == Foo.foo


@pytest.mark.skipif(sys.version_info < (3, 11), reason='StrEnum available after 3.11')
def test_str_enum():
    from enum import StrEnum

    class Foo(StrEnum):
        foo = 'foo'
        bar = 'bar'

    serializer = Serializer(Foo)
    assert serializer.dump(Foo.foo) == 'foo'
    assert serializer.load('bar') == Foo.bar


def test_typed_dict():
    _T = TypeVar('_T')

    class T(TypedDict, Generic[_T]):
        foo_filed: int
        generic_field: _T

    serializer = Serializer(T[bool], camelcase_fields=True)
    assert serializer.dump({'foo_filed': 1, 'generic_field': True}) == {'fooFiled': 1, 'genericField': True}
    assert serializer.load({'fooFiled': 1, 'genericField': True}) == {'foo_filed': 1, 'generic_field': True}


def test_bytes():
    serializer = Serializer(bytes)
    assert serializer.dump(b'foo') == b'foo'
    assert serializer.load(b'foo') == b'foo'



from dataclasses import dataclass
from typing import Annotated, Optional

import pytest
from serpyco_rs import SchemaValidationError, Serializer
from serpyco_rs.metadata import Alias, ForceDefaultForOptional, OmitNone


def test_annotated_filed_alias():
    @dataclass
    class A:
        foo: Annotated[str, Alias('bar')]

    serializer = Serializer(A)

    obj = A(foo='123')

    expected = {'bar': '123'}
    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_omit_none():
    @dataclass
    class A:
        required_val: Optional[bool]
        optional_val: Optional[bool] = None

    serializer = Serializer(Annotated[A, OmitNone])

    expected = {'required_val': None}
    assert serializer.dump(A(required_val=None)) == expected
    assert serializer.load(expected) == A(required_val=None)


@dataclass
class Bar:
    value: Optional[str] = None


@dataclass
class Foo:
    bar: Optional[Bar] = None


@pytest.mark.parametrize(
    ['omit_none', 'foo', 'expected'],
    [
        (True, Foo(), {}),
        (True, Foo(Bar()), {'bar': {}}),
        (False, Foo(), {'bar': None}),
        (False, Foo(Bar()), {'bar': {'value': None}}),
    ],
)
def test_propagete__omit_none(omit_none, foo, expected):
    serializer = Serializer(Foo, omit_none=omit_none)
    res = serializer.dump(foo)

    assert res == expected


def test_omit_none_on_dict():
    serializer = Serializer(dict[str, Optional[bool]], omit_none=True)
    assert serializer.dump({'foo': True, 'bar': None}) == {'foo': True}


def test_force_default_for_optional__propagate_to_nested():
    @dataclass
    class A:
        bar: Optional[bool]

    @dataclass
    class B:
        foo: Optional[A]

    serializer = Serializer(B, force_default_for_optional=True)

    assert serializer.load({}) == B(foo=None)
    assert serializer.load({'foo': {}}) == B(foo=A(bar=None))


def test_force_default_for_optional__override_by_annotated():
    @dataclass
    class A:
        val: Optional[bool]

    @dataclass
    class B:
        foo: Optional[A]
        bar: Annotated[Optional[A], ForceDefaultForOptional]

    serializer = Serializer(B)

    assert serializer.load({'foo': None}) == B(foo=None, bar=None)

    # foo is not annotated, and A.val is still nullable+required
    with pytest.raises(SchemaValidationError):
        serializer.load({'foo': {}})

    # bar is annotated, and A.val is nullable+non required
    assert serializer.load({'foo': None, 'bar': {}}) == B(foo=None, bar=A(val=None))



import sys
from collections.abc import Mapping, Sequence
from dataclasses import dataclass, field
from enum import Enum
from typing import Annotated, Literal, Optional

import attr
import pytest
from serpyco_rs import SchemaValidationError, Serializer
from serpyco_rs.metadata import CamelCase, NoFormat


def test_dump_simple_fields_types():
    @dataclass
    class A:
        int_f: int
        float_f: float
        bool_f: bool
        str_f: str

    serializer = Serializer(A)

    obj = A(
        int_f=123,
        float_f=3.14,
        bool_f=True,
        str_f='Test',
    )
    expected = {'bool_f': True, 'float_f': 3.14, 'int_f': 123, 'str_f': 'Test'}
    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_simple_nested_dataclasses():
    @dataclass
    class C:
        value: int

    @dataclass
    class B:
        value: str
        nested: C

    @dataclass
    class A:
        int_f: int
        nested: B

    serializer = Serializer(A)

    obj = A(
        int_f=123,
        nested=B(value='test', nested=C(value=1)),
    )

    expected = {'int_f': 123, 'nested': {'nested': {'value': 1}, 'value': 'test'}}

    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_iterables():
    @dataclass
    class A:
        iterable_builtins_list: list[int]
        iterable_typing_list: list[int]
        iterable_builtins_sequence: Sequence[int]

    serializer = Serializer(A)

    obj = A(
        iterable_builtins_list=[1, 2, 3],
        iterable_typing_list=[1, 2, 3],
        iterable_builtins_sequence=[1, 2, 3],
    )

    expected = {
        'iterable_builtins_list': [1, 2, 3],
        'iterable_typing_list': [1, 2, 3],
        'iterable_builtins_sequence': [1, 2, 3],
    }

    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_mappings():
    @dataclass
    class B:
        value: str

    @dataclass
    class A:
        dict_field: dict[str, int]
        mapping_field: Mapping[str, B]

    serializer = Serializer(A)

    obj = A(dict_field={'foo': 1}, mapping_field={'bar': B(value='123')})

    expected = {
        'dict_field': {'foo': 1},
        'mapping_field': {'bar': {'value': '123'}},
    }

    assert serializer.load(expected) == obj
    assert serializer.dump(obj) == expected


def test_required_and_nullable():
    @dataclass
    class ReqNotNull:
        foo: int

    @dataclass
    class ReqNullable:
        foo: Optional[int]

    @dataclass
    class OptionalNotNull:
        foo: int = 1

    @dataclass
    class OptionalNullable:
        foo: Optional[int] = 1

    req_not_null = Serializer(ReqNotNull)
    req_nullable = Serializer(ReqNullable)
    optional_not_null = Serializer(OptionalNotNull)
    optional_nullable = Serializer(OptionalNullable)

    assert req_not_null.load({'foo': 2}) == ReqNotNull(foo=2)
    with pytest.raises(SchemaValidationError):
        req_not_null.load({'foo': None})
    with pytest.raises(SchemaValidationError):
        req_not_null.load({})

    assert req_nullable.load({'foo': 2}) == ReqNullable(foo=2)
    assert req_nullable.load({'foo': None}) == ReqNullable(foo=None)
    with pytest.raises(SchemaValidationError):
        req_nullable.load({})

    assert optional_not_null.load({'foo': 2}) == OptionalNotNull(foo=2)
    with pytest.raises(SchemaValidationError):
        assert optional_not_null.load({'foo': None})
    assert optional_not_null.load({}) == OptionalNotNull(foo=1)

    assert optional_nullable.load({'foo': 2}) == OptionalNullable(foo=2)
    assert optional_nullable.load({'foo': None}) == OptionalNullable(foo=None)
    assert optional_nullable.load({}) == OptionalNullable(foo=1)


def test_nullable_without_default():
    @dataclass
    class Entity:
        foo: Optional[int]

    entity_serializer = Serializer(Entity)

    assert entity_serializer.load({'foo': 1}) == Entity(foo=1)
    with pytest.raises(SchemaValidationError):
        entity_serializer.load({})


def test_nullable_without_default__force_none_as_default_for_optional():
    @dataclass
    class Entity:
        foo: Optional[int]

    entity_serializer = Serializer(Entity, force_default_for_optional=True)

    assert entity_serializer.load({'foo': 1}) == Entity(foo=1)
    assert entity_serializer.load({}) == Entity(foo=None)


def test_required__force_none_as_default_for_optional__required_field_not_changed():
    @dataclass
    class Entity:
        foo: int

    entity_serializer = Serializer(Entity, force_default_for_optional=True)

    assert entity_serializer.load({'foo': 1}) == Entity(foo=1)
    with pytest.raises(SchemaValidationError):
        assert entity_serializer.load({})


def test_required_and_nullable_list():
    @dataclass
    class Entity:
        foo: Optional[list[Optional[int]]] = None

    entity_serializer = Serializer(Entity)

    assert entity_serializer.load({}) == Entity(foo=None)
    assert entity_serializer.load({'foo': None}) == Entity(foo=None)
    assert entity_serializer.load({'foo': []}) == Entity(foo=[])
    assert entity_serializer.load({'foo': [1]}) == Entity(foo=[1])
    assert entity_serializer.load({'foo': [1, None]}) == Entity(foo=[1, None])


def test_defaults():
    @dataclass
    class Entity:
        foo: str = '123'
        bar: list[int] = field(default_factory=lambda: list([1, 2, 3]))

    entity_serializer = Serializer(Entity)

    assert entity_serializer.load({'bar': [1]}) == Entity(foo='123', bar=[1])
    assert entity_serializer.load({}) == Entity(foo='123', bar=[1, 2, 3])


if sys.version_info >= (3, 10):

    def test_union_optional__dump_load__ok():
        # arrange
        @dataclass
        class UnionClass:
            name: str | None
            count: None | int

        # act
        serializer = Serializer(UnionClass)

        # assert
        foo = UnionClass(name=None, count=None)
        dict_foo = {'name': None, 'count': None}
        assert serializer.dump(foo) == dict_foo
        assert foo == serializer.load(dict_foo)

        bar = UnionClass(name='try', count=5)
        dict_bar = {'name': 'try', 'count': 5}
        assert serializer.dump(bar) == dict_bar
        assert bar == serializer.load(dict_bar)


def test_serializer_with_camelcase():
    @dataclass
    class C:
        foo_field: int

    @dataclass
    class B:
        some_value: str
        another_value: Annotated[C, CamelCase]

    @dataclass
    class A:
        dict_field: dict[str, int]
        inner_value_one: B
        inner_value_two: Annotated[B, NoFormat]

    serializer = Serializer(A, camelcase_fields=True)

    obj = A(
        dict_field={'foo': 1},
        inner_value_one=B(some_value='123', another_value=C(11)),
        inner_value_two=B(some_value='1', another_value=C(22)),
    )

    expected = {
        'dictField': {'foo': 1},
        'innerValueOne': {'someValue': '123', 'anotherValue': {'fooField': 11}},
        'innerValueTwo': {'some_value': '1', 'another_value': {'fooField': 22}},
    }

    assert serializer.load(expected) == obj
    assert serializer.dump(obj) == expected


def test_load_frozen_dataclass():
    @dataclass(frozen=True)
    class Foo:
        bar: bool

    serializer = Serializer(Foo)

    assert serializer.load({'bar': True}) == Foo(bar=True)


def test_load_frozen_attrs():
    @attr.frozen
    class Foo:
        bar: bool

    serializer = Serializer(Foo)

    assert serializer.load({'bar': True}) == Foo(bar=True)


def test_load_literal():
    class FooEnum(Enum):
        three = '3'

    @dataclass
    class Foo:
        bar: Literal[1, '2', FooEnum.three]

    serializer = Serializer(Foo)

    assert serializer.load({'bar': 1}) == Foo(bar=1)
    assert serializer.load({'bar': '2'}) == Foo(bar='2')
    assert serializer.load({'bar': '3'}) == Foo(bar=FooEnum.three)


def test_dump_literal():
    class FooEnum(Enum):
        three = '3'

    @dataclass
    class Foo:
        bar: Literal[1, '2', FooEnum.three]

    serializer = Serializer(Foo)

    assert serializer.dump(Foo(bar=1)) == {'bar': 1}
    assert serializer.dump(Foo(bar='2')) == {'bar': '2'}
    assert serializer.dump(Foo(bar=FooEnum.three)) == {'bar': '3'}


def test_load_new_type():
    from typing import NewType

    Foo = NewType('Foo', int)

    @dataclass
    class Bar:
        foo: Foo

    serializer = Serializer(Bar)

    assert serializer.load({'foo': 1}) == Bar(foo=Foo(1))


def test_dump_new_type():
    from typing import NewType

    Foo = NewType('Foo', int)

    @dataclass
    class Bar:
        foo: Foo

    serializer = Serializer(Bar)

    assert serializer.dump(Bar(foo=Foo(1))) == {'foo': 1}



from ._custom_types import CustomType
from ._json_schema import JsonSchemaBuilder
from ._main import Serializer
from .exceptions import ErrorItem, SchemaValidationError, ValidationError


__all__ = ['CustomType', 'ErrorItem', 'JsonSchemaBuilder', 'SchemaValidationError', 'Serializer', 'ValidationError']



from typing import Any

import serpyco_rs

from .base import Dataclass, test_object


_serializer = serpyco_rs.Serializer(Dataclass)

test_object = test_object


def load(data: dict[str, Any]) -> Dataclass:
    return _serializer.load(data)


def dump(obj: Dataclass) -> dict[str, Any]:
    return _serializer.dump(obj)



from typing import Generic, TypeVar

from ._serpyco_rs import (
    AnyType,
    ArrayType,
    BaseType,
    BooleanType,
    BytesType,
    CustomEncoder as _CustomEncoder,
    DateTimeType,
    DateType,
    DecimalType,
    DefaultValue as _DefaultValue,
    DictionaryType,
    DiscriminatedUnionType,
    EntityField,
    EntityType,
    EnumType,
    ErrorItem,
    FloatType,
    IntegerType,
    LiteralType,
    OptionalType,
    RecursionHolder,
    SchemaValidationError,
    Serializer,
    StringType,
    TimeType,
    TupleType,
    TypedDictType,
    UnionType,
    UUIDType,
    ValidationError,
    CustomType,
)

__all__ = [
    'AnyType',
    'ArrayType',
    'BaseType',
    'BooleanType',
    'BytesType',
    'CustomEncoder',
    'CustomType',
    'DateTimeType',
    'DateType',
    'DecimalType',
    'DefaultValue',
    'DictionaryType',
    'DiscriminatedUnionType',
    'EntityField',
    'EntityType',
    'EnumType',
    'ErrorItem',
    'FloatType',
    'IntegerType',
    'LiteralType',
    'OptionalType',
    'RecursionHolder',
    'SchemaValidationError',
    'Serializer',
    'StringType',
    'TimeType',
    'TupleType',
    'TypedDictType',
    'UUIDType',
    'UnionType',
    'ValidationError',
]


_T = TypeVar('_T')
_I = TypeVar('_I')
_O = TypeVar('_O')


class CustomEncoder(_CustomEncoder, Generic[_I, _O]):
    """pyo3 doesn't support specifying concrete types for generic methods."""


class DefaultValue(_DefaultValue, Generic[_T]):
    """pyo3 doesn't support specifying concrete types for generic methods."""


NOT_SET = DefaultValue.none()



from dataclasses import dataclass
from typing import Annotated, Generic, Literal, TypeVar, Optional, Union

import pytest
import sys

from serpyco_rs import Serializer
from serpyco_rs.metadata import Discriminator


TWidget_co = TypeVar('TWidget_co', bound='BaseWidget', covariant=True)


@dataclass
class BaseWidget(Generic[TWidget_co]):
    type: str
    childrens: Optional[list[TWidget_co]] = None


@dataclass
class Widget1(BaseWidget[TWidget_co]):
    type: Literal['Widget1'] = 'Widget1'


@dataclass
class Widget2(BaseWidget[TWidget_co]):
    type: Literal['Widget2'] = 'Widget2'
    some_field: Optional[str] = None


Widget = Annotated[Union[Widget1['Widget'], Widget2['Widget']], Discriminator('type')]


@pytest.mark.skipif(sys.version_info < (3, 11), reason='requires python3.11 or higher')
def test_recursive_generics():
    serializer = Serializer(Widget)

    obj = Widget1(type='Widget1', childrens=[Widget2(type='Widget2', some_field='some_value')])

    data = {'type': 'Widget1', 'childrens': [{'type': 'Widget2', 'some_field': 'some_value', 'childrens': None}]}

    assert serializer.dump(obj) == data
    assert serializer.load(data) == obj


@pytest.mark.skipif(sys.version_info < (3, 11), reason='requires python3.11 or higher')
def test_recursive_generics_propagates_annotations():
    serializer = Serializer(Widget, camelcase_fields=True)

    obj = Widget1(type='Widget1', childrens=[Widget2(type='Widget2', some_field='some_value', childrens=[])])

    data = {'type': 'Widget1', 'childrens': [{'type': 'Widget2', 'someField': 'some_value', 'childrens': []}]}

    assert serializer.dump(obj) == data
    assert serializer.load(data) == obj



from dataclasses import dataclass
from typing import Any, Optional, TypeVar, overload

from typing_extensions import Self

from serpyco_rs._impl import BaseType
from serpyco_rs.metadata import FieldFormat, NoneAsDefaultForOptional, NoneFormat


@dataclass
class ResolverContext:
    """Context for the type resolution process"""

    globals: dict[str, Any]
    type_cache: dict[str, Optional[BaseType]]
    discriminator_field: Optional[str] = None

    def cache_type(self, key: str, value: Optional[BaseType]) -> None:
        """Cache a resolved type"""
        self.type_cache[key] = value

    def __getitem__(self, item: str) -> Optional[BaseType]:
        return self.type_cache.get(item)

    def get_cached_type(self, key: str) -> Optional[BaseType]:
        """Get a cached type by key"""
        return self.type_cache.get(key)

    def has_cached_type(self, key: str) -> bool:
        """Check if a type is cached"""
        return key in self.type_cache


_T = TypeVar('_T')

_PROPAGETABLE_METADATA_TYPES = {FieldFormat, NoneAsDefaultForOptional, NoneFormat}


class Annotations:
    """Storage for field annotations from typing.Annotated"""

    _data: dict[type, Any]

    def __init__(self, *args: Any):
        self._data = {}
        for arg in args:
            self._data[type(arg)] = arg

    @overload
    def get(self, key: type[_T], default: None = None) -> Optional[_T]: ...

    @overload
    def get(self, key: type[_T], default: _T) -> _T: ...

    def get(self, key: type[_T], default: Optional[_T] = None) -> Optional[_T]:
        return self._data.get(key, default)

    def merge(self, other: Self) -> Self:
        new_data = self.__class__()
        new_data._data = {
            **{k: v for k, v in self._data.items() if k in _PROPAGETABLE_METADATA_TYPES},
            **other._data,
        }
        return new_data

    def make_key(self) -> str:
        d = list(map(str, self._data.values()))
        d.sort()
        return str(d)

    def __eq__(self, value: object, /) -> bool:
        if isinstance(value, self.__class__):
            return self._data == value._data
        return False

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({tuple(self._data.values())})'



import os
import sys

import nox


nox.options.sessions = ['test', 'lint', 'type_check', 'rust_lint']


def build(session, use_pip: bool = False):
    if _is_ci():
        # Install form wheels
        install(session, '--no-index', '--no-deps', '--find-links', 'wheels/', 'serpyco-rs', use_pip=use_pip)
        install(session, '--find-links', 'wheels/', 'serpyco-rs', use_pip=use_pip)
        return

    session.run_always('maturin', 'develop', '-r')


@nox.session(python=False)
def test(session):
    build(session)
    install(session, '-r', 'requirements/dev.txt')
    session.run('pytest', '-vvs', 'tests/', *session.posargs)


@nox.session(python=False)
def lint(session):
    build(session)
    install(session, '-r', 'requirements/lint.txt')

    session.cd('python/serpyco_rs')
    paths = ['.', '../../tests', '../../bench']
    session.run('ruff', 'format',  *(['--check', '--diff', *paths] if _is_ci() else paths))
    session.run('ruff', 'check', '.', *([] if _is_ci() else ['--fix']))


@nox.session(python=False)
def rust_lint(session):
    session.run('cargo', 'fmt', '--all', *(['--', '--check'] if _is_ci() else []))
    session.run('cargo', 'clippy', '--all-targets', '--all-features', '--', '-D', 'warnings')


@nox.session(python=False)
def type_check(session):
    build(session)
    install(session, '-r', 'requirements/type_check.txt')

    session.cd('python/serpyco_rs')
    session.run('pyright', '.', success_codes=[0, 1] if _is_ci() else [0])
    session.run('pyright', '.', '--verifytypes', 'serpyco_rs')
    session.run('mypy', '.', '--strict', '--implicit-reexport', '--pretty')


@nox.session(python=False)
def bench(session):
    build(session)
    install(session, '-r', 'requirements/bench.txt')

    session.run(
        'pytest',
        *(session.posargs if session.posargs else ['bench']),
        '--verbose',
        '--benchmark-min-time=0.5',
        '--benchmark-max-time=1',
        '--benchmark-disable-gc',
        '--benchmark-autosave',
        '--benchmark-save-data',
        '--benchmark-compare',
    )


@nox.session(python=False)
def test_rc_leaks(session):
    # uv don't resolve wheels when used python debug build
    build(session, use_pip=True)
    install(session, '-r', 'requirements/bench.txt', use_pip=True)
    session.run(
        'pytest',
        *(session.posargs if session.posargs else ['bench']),
        '--verbose',
        '--debug-refs',
        '--debug-refs-gc',
    )


@nox.session(python=False)
def bench_codespeed(session):
    build(session)
    install(session, '-r', 'requirements/bench.txt')
    install(session, 'pytest-codspeed')
    session.run('pytest', 'bench', '--ignore=bench/compare/test_benchmarks.py', '--codspeed')


def install(session, *args, use_pip: bool = False):
    if session._runner.global_config.no_install:
        return
    cmd = ['pip', 'install'] if use_pip else ['uv', 'pip', 'install', '--python', sys.executable]
    session.run_always(*cmd, *args)


def _is_ci() -> bool:
    return bool(os.environ.get('CI', None))
